package org.amse.vbut.vzab.io.impl;

import org.amse.vbut.vzab.*;
import org.amse.vbut.vzab.io.*;
import org.amse.vbut.vzab.model.*;
import org.amse.vbut.vzab.model.impl.*;

import java.io.*;
import java.util.*;

import org.xml.sax.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 * DOM implementation of the IReader interface
 * 
 * @author But Vasiliy
 *
 */
class DOMReader implements IReader {
	
	public IGame readGame(String filename, int minwidth, int minheight, int maxwidth, int maxheight) {
		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			
			dbf.setIgnoringElementContentWhitespace(true);
			dbf.setIgnoringComments(true);
			
			DocumentBuilder builder = dbf.newDocumentBuilder();
			
			Document document = builder.parse(filename);

			return createGame(document, minwidth, minheight, maxwidth, maxheight);
		} catch (SAXException e) {
			throw new VZabException("Cannot parse file \"" + filename + "\": " + e.getMessage());
		} catch (IOException e) {
			throw new VZabException("Cannot read file \"" + filename + "\": " + e.getMessage());
		} catch (ParserConfigurationException e) {
			throw new VZabException("Cannot create parser \"" + filename + "\": " + e.getMessage());
		}
	}
	
	private IGame createGame(Document document, int minwidth, int minheight, int maxwidth, int maxheight) {
		Element root = document.getDocumentElement();
		
		NodeList nodelist = root.getElementsByTagName("players"); 

		if (nodelist.getLength() > 1) {
			throw new VZabException("There are more than one \"players\" tags");
		} else if (nodelist.getLength() == 0) {
			throw new VZabException("There are no \"players\" tags");
		}
		
		java.util.List<IPlayer> players = readPlayers(nodelist.item(0));
		
		nodelist = root.getElementsByTagName("board");

		if (nodelist.getLength() > 1) {
			throw new VZabException("There are more than one \"board\" tags");
		} else if (nodelist.getLength() == 0) {
			throw new VZabException("There are no \"board\" tags");
		}
		
		return readBoard(nodelist.item(0), players, minwidth, minheight, maxwidth, maxheight);
	}
	
	private java.util.List<IPlayer> readPlayers(Node nodePlayers) {
		if (nodePlayers == null) {
			throw new VZabException("Can not find node \"players\"");
		}
		
		final String nodeName = nodePlayers.getNodeName();

		if (!nodeName.equalsIgnoreCase("players")) {
			throw new VZabException("Wrong node \"" + nodeName + "\". Must be node \"players\"");
		}

		if (!nodePlayers.hasAttributes()) {
			throw new VZabException("Can not find attribute \"number\" in node \"players\"");
		}

		final NamedNodeMap attributes = nodePlayers.getAttributes();
		final Node nodeNumber = attributes.getNamedItem("number");
		if (nodeNumber == null) {
			throw new VZabException("Can not find attribute \"number\" in node \"players\"");
		}
		int number;
		try {
			number = Integer.parseInt(nodeNumber.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"number\" in node \"players\" has invalid format");
		}
		
		if (number <= 0) {
			throw new VZabException("Attribute \"number\" value in node \"players\" must be greater than zero");
		}

		if (!nodePlayers.hasChildNodes()) {
			throw new VZabException("Can not find information about players in players child nodes");
		}
		
		if (nodePlayers.getNodeType() != Node.ELEMENT_NODE) {
			throw new VZabException("Wrong type of <players> node: must be ELEMENT_NODE.");
		}
		
		return parsePlayerNodeList(((Element)nodePlayers).getElementsByTagName("player"), number);
	}

	private IGame readBoard(Node nodeBoard, java.util.List<IPlayer> players, int minwidth, int minheight, int maxwidth, int maxheight) {
		if (nodeBoard == null) {
			throw new VZabException("Can not find node \"board\" in the game");
		}

		final String nodeName = nodeBoard.getNodeName();
		
		if (!nodeName.equalsIgnoreCase("board")) {
			throw new VZabException("Wrong node \"" + nodeName + "\". Must be node \"board\"");
		}
		
		if (!nodeBoard.hasAttributes()) {
			throw new VZabException("Can not find board attributes: width and height");
		}
		
		NamedNodeMap attributes = nodeBoard.getAttributes();
		
		Node nodeWidth = attributes.getNamedItem("width");
		Node nodeHeight = attributes.getNamedItem("height");
		
		if (nodeWidth == null) {
			throw new VZabException("Can not find attribute \"width\" in the node \"board\"");
		}
		
		if (nodeHeight == null) {
			throw new VZabException("Can not find attribute \"height\" in the node \"board\"");
		}
		
		int width;
		try {
			width = Integer.parseInt(nodeWidth.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"width\" in node \"board\" has invalid format");
		}

		int height;
		try {
			height = Integer.parseInt(nodeHeight.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"height\" in node \"board\" has invalid format");
		}

		if (width < minwidth || width > maxwidth
				|| height < minheight || height > maxheight) {
			throw new VZabException("Unsupported board dimensions");
		}
		
		IGame game = ModelFactory.newGame(width, height, players);
		
		if (nodeBoard.getNodeType() != Node.ELEMENT_NODE) {
			throw new VZabException("Wrong type of <board> node: must be ELEMENT_NODE.");
		}
		
		Element elementBoard = (Element)nodeBoard;

		NodeList moves = elementBoard.getElementsByTagName("moves");
		if (moves.getLength() == 0) {
			throw new VZabException("Can not find moves");
		} else if (moves.getLength() > 1) {
			throw new VZabException("Too many <moves> nodes");
		}
		
		parseBoardMoves(game, moves.item(0));
		
		return game;
	}
	
	private void parseBoardMoves(IGame game, Node nodeMoves) {
		
		if (!nodeMoves.hasAttributes()) {
			throw new VZabException("Can not find attributes of the node \"moves\"");
		}
		
		NamedNodeMap attributes = nodeMoves.getAttributes();

		Node nodeNumber = attributes.getNamedItem("number");
		if (nodeNumber == null) {
			throw new VZabException("Node \"moves\" doesn't contain attribute \"number\"");
		}
		int number;
		try {
			number = Integer.parseInt(nodeNumber.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"number\" in node \"moves\" has invalid format");
		}
		
		if (number == 0) {
			return;
		} else if (number < 0) {
			throw new VZabException("Attribute \"number\" value must be greater or equal to zero");
		}
		
		Node nodeCurrentMove = attributes.getNamedItem("currentmove");
		if (nodeCurrentMove == null) {
			throw new VZabException("Can not find attribute \"currentmove\"");
		}
		int currentmove;
		try {
			currentmove = Integer.parseInt(nodeCurrentMove.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"currentmove\" in node \"moves\" has invalid format");
		}
		
		if (nodeMoves.getNodeType() != Node.ELEMENT_NODE) {
			throw new VZabException("Wrong type of <moves> node: must be ELEMENT_NODE.");
		}
		
		if (currentmove < 0 || currentmove > number) {
			throw new VZabException("Wrong value of current move");
		}

		NodeList moves = ((Element)nodeMoves).getElementsByTagName("move");
		
		if (moves.getLength() < number) {
			throw new VZabException("There are not enough moves: must be " + number + " moves; found " + moves.getLength());
		} else if (moves.getLength() > number) {
			throw new VZabException("There are too many moves: must be " + number + " moves; found " + moves.getLength());
		}

		final java.util.List<IPlayer> players = game.getPlayers();
		final IBoard board = game.getBoard();
		
		for (int i = 0; i < number; ++i) {
			Node move = findMove(moves, i);
			
			if (move == null) {
				throw new VZabException("Can not find move number " + i);
			}
			
			if (!move.hasAttributes()) {
				throw new VZabException("Can not find attributes of the node \"move\"");
			}
			
			NamedNodeMap moveattributes = move.getAttributes();
			
			Node nodeX, nodeY, nodePlayer;
			
			if ((nodeX = moveattributes.getNamedItem("x")) == null) {
				throw new VZabException("Can not find attribute \"x\"");
			}
			
			if ((nodeY = moveattributes.getNamedItem("y")) == null) {
				throw new VZabException("Can not find attribute \"y\"");
			}
			
			if ((nodePlayer = moveattributes.getNamedItem("player")) == null) {
				throw new VZabException("Can not find attribute \"player\"");
			}
			
			int x;
			try {
				x= Integer.parseInt(nodeX.getNodeValue());
			} catch (NumberFormatException e) {
				throw new VZabException("Attribute \"x\" in node \"move\" with index = " + i + " has invalid format");
			}
			
			int y;
			try {
				y = Integer.parseInt(nodeY.getNodeValue());
			} catch (NumberFormatException e) {
				throw new VZabException("Attribute \"y\" in node \"move\" with index = " + i + " has invalid format");
			}
			
			int playernum;
			try {
				playernum = Integer.parseInt(nodePlayer.getNodeValue());
			} catch (NumberFormatException e) {
				throw new VZabException("Attribute \"playernum\" in node \"move\" with index = " + i + " has invalid format");
			}
			
			if (x < 0 || x >= board.getWidth()) {
				throw new VZabException("Node \"move\" with index " + i + " has invalid coordinate x=" + x);
			}

			if (y < 0 || y >= board.getHeight()) {
				throw new VZabException("Node \"move\" with index " + i + " has invalid coordinate y=" + y);
			}

			if (playernum < 0 || playernum >= players.size()) {
				throw new VZabException("Node \"move\" with index " + i + " has invalid value of attribute \"player\"");
			}
			
			IPlayer player = players.get(playernum);
			
			BoardPoint bp = game.getParentPoint(x, y, player);
			
			if (bp == null) {
				throw new VZabException("Move number " + i + " is bad");
			}
			
			final IPiece parent = board.getPiece(bp.getX(), bp.getY());
			
			if (!game.canDoMove(x, y, parent, true))
			{
				throw new VZabException("Move number " + i + " is bad");
			}
			
			game.doMove(x, y, parent);
		}
		
		while (game.getCurrentMove() > currentmove)
		{
			game.undoMove();
		}

		if (!game.setCurrentMove(currentmove))
		{
			throw new VZabException("Unable to set current move = " + currentmove);
		}
	}
	
	private Node findMove(NodeList moves, int index) {
		final int size = moves.getLength();
		for (int i = 0; i < size; ++i) {
			
			final Node move = moves.item(i);
			
			final NamedNodeMap attributes = move.getAttributes();
			
			if (attributes == null) {
				continue;
			}
			
			final Node nodeIndex = attributes.getNamedItem("index");
			
			if (nodeIndex == null) {
				continue;
			}
			
			int value;
			try {
				value = Integer.parseInt(nodeIndex.getNodeValue());
			} catch (NumberFormatException e) {
				throw new VZabException("Attribute \"index\" in node \"move\" has invalid format value = " + nodeIndex.getNodeValue());
			}
			
			if (value == index) {
				return move;
			}
		}
		return null;
	}
	
	private List<IPlayer> parsePlayerNodeList(NodeList list, int number) {
		final IPlayer[] players = new IPlayer[number];

		final IPlayerIO pr = IOFactory.newPlayerIO();
		
		if (list.getLength() != number) {
			throw new VZabException("List of players doesn't contain " + number + " elements");
		}
		
		for (int i = 0; i < number; ++i) {
			Node node = list.item(i);

			if (!node.hasAttributes()) {
				throw new VZabException("Node \"player\" doesn't contain required attributes: \"filename\", \"name\" and \"index\"");
			}
			
			NamedNodeMap attributes = node.getAttributes();
			
			Node nodePlayerName = attributes.getNamedItem("name");
			if (nodePlayerName == null) {
				throw new VZabException("Node \"player\" doesn't contain attribute \"name\"");
			}
			
			/*Node nodePlayerColor = attributes.getNamedItem("color");
			if (nodePlayerColor == null) {
				throw new VZabException("Node \"player\" doesn't contain attribute \"color\"");
			}*/
			
			Node nodeFilename = attributes.getNamedItem("filename");
			if (nodeFilename == null) {
				throw new VZabException("Node \"player\" doesn't contain attribute \"filename\"");
			}
			
			boolean isComputer = false;
			Node nodeIsComputer = attributes.getNamedItem("iscomputer"); 
			if (nodeIsComputer != null) {
				isComputer = Boolean.parseBoolean(nodeIsComputer.getNodeValue());
			}
		
			/*IPlayer player = ModelFactory.newPlayer(nodePlayerName.getNodeValue(),
					new VZabColor(nodePlayerColor.getNodeValue(), 16), isComputer);*/
			
			IPlayer player = pr.readPlayer(nodeFilename.getNodeValue(), 
					nodePlayerName.getNodeValue());
			
			player.setComputer(isComputer);
			
			Node nodeIndex = attributes.getNamedItem("index");
			if (nodeIndex == null) {
				throw new VZabException("Node \"player\" doesn't contain attribute \"index\"");
			}

			int index;
			try {
				index = Integer.parseInt(nodeIndex.getNodeValue()); 
			} catch (NumberFormatException e) {
				throw new VZabException("Attribute \"index\" in node \"player\" has invalid format value = " + nodeIndex.getNodeValue());
			} 
				
			if (index < 0 || index >= players.length) {
				throw new VZabException("Node \"player\" has invalid value of attribute \"index\"");
			}
			
			if (players[index] != null) {
				throw new VZabException("Duplicating player entry with index " + index);
			}
			players[index] = player;
		}
		
		java.util.LinkedList<IPlayer> playerslist = new java.util.LinkedList<IPlayer>();
		
		if (!Collections.addAll(playerslist, players)) {
			throw new VZabException("Can't create players list");
		}
		
		return playerslist;
	}

	public List<IPlayer> readPlayers(String filename) {
		Document document;
		
		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			
			dbf.setIgnoringElementContentWhitespace(true);
			dbf.setIgnoringComments(true);
			
			DocumentBuilder builder = dbf.newDocumentBuilder();
			
			document = builder.parse(filename);
			
		} catch (SAXException e) {
			throw new VZabException("Cannot parse file \"" + filename + "\": " + e.getMessage());
		} catch (IOException e) {
			throw new VZabException("Cannot read file \"" + filename + "\": " + e.getMessage());
		} catch (ParserConfigurationException e) {
			throw new VZabException("Cannot create parser \"" + filename + "\": " + e.getMessage());
		}

		final Element root = document.getDocumentElement();
		
		if (!root.getNodeName().equalsIgnoreCase("players")) {
			throw new VZabException("Main tag must be \"players\" tag");
		}
		
		return readPlayers(root);
	}

	private class Settings implements ISettings {

		private int myMainX; 
		private int myMainY;

		private boolean myDieMessageShow;
		private boolean myGrayMessageShow;
		private boolean myWinMessageShow;
		
		public Settings(int mainx, int mainy, 
				boolean dieshow, boolean grayshow, boolean winshow) {
			myMainX = mainx;
			myMainY = mainy;
			
			myDieMessageShow = dieshow;
			myGrayMessageShow = grayshow;
			myWinMessageShow = winshow;
		}
		
		public int getMainWindowX() {
			return myMainX;
		}

		public int getMainWindowY() {
			return myMainY;
		}

		public boolean isWinMessageShow() {
			return myWinMessageShow;
		}

		public boolean isDieMessageShow() {
			return myDieMessageShow;
		}

		public boolean isGrayMessageShow() {
			return myGrayMessageShow;
		}
	}
	
	public ISettings readSettings(String filename) {
		Document document;
		
		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			
			dbf.setIgnoringElementContentWhitespace(true);
			dbf.setIgnoringComments(true);
			
			DocumentBuilder builder = dbf.newDocumentBuilder();
			
			document = builder.parse(filename);
			
		} catch (SAXException e) {
			throw new VZabException("Cannot parse file \"" + filename + "\": " + e.getMessage());
		} catch (IOException e) {
			throw new VZabException("Cannot read file \"" + filename + "\": " + e.getMessage());
		} catch (ParserConfigurationException e) {
			throw new VZabException("Cannot create parser \"" + filename + "\": " + e.getMessage());
		}

		final Element root = document.getDocumentElement();

		return readSettings(root);
	}
	
	private ISettings readSettings(Element root) {
		
		if (!root.getNodeName().equals("settings")) {
			throw new VZabException("Main node must be node \"settings\"");
		}
		
		NodeList positions = root.getElementsByTagName("position");
		if (positions.getLength() == 0) {
			throw new VZabException("There is no \"position\" node: position of the main window");
		} else if (positions.getLength() > 1) {
			throw new VZabException("There is too many \"position\" nodes");
		}
		BoardPoint bp = readPosition(positions.item(0));
		
		NodeList dialogs = root.getElementsByTagName("dialogs");
		if (dialogs.getLength() == 0) {
			throw new VZabException("There is no \"dialogs\" node: dialogs settings");
		} else if (dialogs.getLength() > 1) {
			throw new VZabException("There is too many \"dialogs\" nodes");
		}
		Node dialogsnode = dialogs.item(0);
		
		return new Settings(bp.getX(), bp.getY(), 
				readDieShow(dialogsnode), 
				readGrayShow(dialogsnode), 
				readWinShow(dialogsnode));
	}
	
	private boolean readDieShow(Node dialogs) {
		if (!dialogs.hasAttributes()) {
			throw new VZabException("Node \"dialogs\" doesn't contain attributes");
		}
		NamedNodeMap attributes = dialogs.getAttributes();
		Node nodeShow;
		if ((nodeShow = attributes.getNamedItem("dieshow")) == null) {
			throw new VZabException("Node \"dialogs\" doesn't contain attribute \"dieshow\"");
		}
		return Boolean.parseBoolean(nodeShow.getNodeValue());
	}
	
	private boolean readWinShow(Node dialogs) {
		if (!dialogs.hasAttributes()) {
			throw new VZabException("Node \"dialogs\" doesn't contain attributes");
		}
		NamedNodeMap attributes = dialogs.getAttributes();
		Node nodeShow;
		if ((nodeShow = attributes.getNamedItem("winshow")) == null) {
			throw new VZabException("Node \"dialogs\" doesn't contain attribute \"winshow\"");
		}
		return Boolean.parseBoolean(nodeShow.getNodeValue());
	}
	
	private boolean readGrayShow(Node dialogs) {
		if (!dialogs.hasAttributes()) {
			throw new VZabException("Node \"dialogs\" doesn't contain attributes");
		}
		NamedNodeMap attributes = dialogs.getAttributes();
		Node nodeShow;
		if ((nodeShow = attributes.getNamedItem("grayshow")) == null) {
			throw new VZabException("Node \"dialogs\" doesn't contain attribute \"grayshow\"");
		}
		return Boolean.parseBoolean(nodeShow.getNodeValue());
	}
	
	private BoardPoint readPosition(Node position) {
		if (!position.hasAttributes()) {
			throw new VZabException("Node \"position\" doesn't contain attributes \"x\" and \"y\"");
		}
		
		NamedNodeMap attributes = position.getAttributes();

		Node nodeX;
		Node nodeY;

		if ((nodeX = attributes.getNamedItem("x")) == null) {
			throw new VZabException("Node \"positions\" doesn't contain attribute \"x\"");
		}
		if ((nodeY = attributes.getNamedItem("y")) == null) {
			throw new VZabException("Node \"positions\" doesn't contain attribute \"y\"");
		}
		
		int x;
		try {
			x = Integer.parseInt(nodeX.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"x\" in node \"position\" has invalid format");
		}
		
		int y;
		try {
			y = Integer.parseInt(nodeY.getNodeValue());
		} catch (NumberFormatException e) {
			throw new VZabException("Attribute \"y\" in node \"position\" has invalid format");
		}

		return new BoardPoint(x, y);
	}

	public ISettings getDefaultSettings() {
		return new Settings(0, 0, true, true, true);
	}

	public String readReadme(String filename) {
		Document document;
		
		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			
			dbf.setIgnoringElementContentWhitespace(true);
			dbf.setIgnoringComments(true);
			
			DocumentBuilder builder = dbf.newDocumentBuilder();
			
			if (filename.charAt(0) != '/') {
				filename = '/' + filename;
			}
			InputStream is = this.getClass().getResourceAsStream(filename);
			
			document = builder.parse(is);
			
		} catch (SAXException e) {
			throw new VZabException("Cannot parse file \"" + filename + "\": " + e.getMessage());
		} catch (IOException e) {
			throw new VZabException("Cannot read file \"" + filename + "\": " + e.getMessage());
		} catch (ParserConfigurationException e) {
			throw new VZabException("Cannot create parser \"" + filename + "\": " + e.getMessage());
		}

		final Element root = document.getDocumentElement();

		if (!root.getNodeName().equals("readme")) {
			throw new VZabException("Can't load readme file \"" + filename + "\"");
		}
		
		return root.getTextContent();
	}
}
