package org.amse.vbut.vzab.io.impl;

import org.amse.vbut.vzab.io.*;
import org.amse.vbut.vzab.model.*;
import org.amse.vbut.vzab.*;

import java.io.*;
import java.util.List;

import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 * DOM implementation of the IWriter class
 * 
 * @author But Vasiliy
 *
 */
class DOMWriter implements IWriter {

	public void writeGame(IGame game, String filename) {
		
		if (game == null) {
			throw new VZabException("There is no game to save");
		}
		
		final DocumentBuilderFactory drf = DocumentBuilderFactory.newInstance();
		
		DocumentBuilder builder;
		try {
			builder = drf.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new VZabException(e.getMessage());
		}

		final Document document = builder.newDocument();
		
		fillDOMDocument(game, document);
		
		saveToFile(document, filename);
	}
	
	private void fillDOMDocument(IGame game, Document document) {
		final Element elementGame = document.createElement("game");
		
		elementGame.appendChild(createPlayersElement(game.getPlayers(), document));
		elementGame.appendChild(createBoardElement(game, document));
		
		document.appendChild(elementGame);
	}
	
	private Element createPlayersElement(java.util.List<IPlayer> players, Document document) {
		final Element elementPlayers = document.createElement("players");
		
		final int size = players.size();
		
		elementPlayers.setAttribute("number", String.valueOf(size));
		
		for (int i = 0; i < size; ++i) {
			final IPlayer player = players.get(i);

			Element elementPlayer = document.createElement("player");
			
			elementPlayer.setAttribute("index", String.valueOf(i));
			elementPlayer.setAttribute("name", player.getName());
			elementPlayer.setAttribute("filename", player.getPath());
			
			if (player.isComputer()) {
				elementPlayer.setAttribute("iscomputer", "true");
			}
			
			elementPlayers.appendChild(elementPlayer);
		}

		return elementPlayers;
	}
	
	private Element createBoardElement(IGame game, Document document) {
		final IBoard board = game.getBoard();
		final Element elementBoard = document.createElement("board");
		
		elementBoard.setAttribute("width", String.valueOf(board.getWidth()));
		elementBoard.setAttribute("height", String.valueOf(board.getHeight()));
		
		elementBoard.appendChild(createMovesElement(game, document));

		return elementBoard;
	}

	private Element createMovesElement(IGame game, Document document) {
		final Element elementMoves = document.createElement("moves");
		
		final java.util.List<IMove> moves = game.getMoves();
		final int size = moves.size();
		
		elementMoves.setAttribute("number", String.valueOf(size));
		
		if (size == 0) {
			return elementMoves;
		}
		
		elementMoves.setAttribute("currentmove", String.valueOf(game.getCurrentMove()));
		
		for (int i = 0; i < size; ++i) {
			IMove move = moves.get(i);
			
			Element elementMove = document.createElement("move");
			
			elementMove.setAttribute("index", String.valueOf(i));
			elementMove.setAttribute("player", String.valueOf(game.getPlayerIndex(move.getPiece().getPlayer())));
			elementMove.setAttribute("x", String.valueOf(move.getX()));
			elementMove.setAttribute("y", String.valueOf(move.getY()));
			
			elementMoves.appendChild(elementMove);
		}
		
		return elementMoves;
	}
	
	private void saveToFile(Document doc, String filename) {
		Source source = new DOMSource(doc);
		
		File file = new File(filename);
		
		try {
			Result result = new StreamResult( new FileWriter(file) );
			
			TransformerFactory factory = TransformerFactory.newInstance();
			
			factory.setAttribute("indent-number", new Integer(4));
			
			Transformer xform = factory.newTransformer();
			
			xform.setOutputProperty(OutputKeys.INDENT, "yes");
			
			xform.transform(source, result);
		} catch (TransformerConfigurationException e) {
			throw new VZabException(e.getMessage());
		} catch (TransformerException e) {
			throw new VZabException(e.getMessage());
		} catch (IOException e) {
			throw new VZabException(e.getMessage());
		}
	}

	public void writePlayers(List<IPlayer> players, String filename) {
		if (players == null) {
			throw new VZabException("There is no player list to save");
		}
		
		final DocumentBuilderFactory drf = DocumentBuilderFactory.newInstance();
		
		DocumentBuilder builder;
		try {
			builder = drf.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new VZabException(e.getMessage());
		}

		final Document document = builder.newDocument();
		
		document.appendChild( createPlayersElement(players, document) );
		
		saveToFile(document, filename);
	}

	public void writeSettings(String filename, int mainx, int mainy, 
			boolean dieshow, boolean grayshow, boolean winshow) {
		final DocumentBuilderFactory drf = DocumentBuilderFactory.newInstance();
		
		DocumentBuilder builder;
		try {
			builder = drf.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new VZabException(e.getMessage());
		}

		final Document document = builder.newDocument();
		
		document.appendChild( createSettingsElement(document, mainx, mainy, dieshow, grayshow, winshow) );
		
		saveToFile(document, filename);
		
		//rewriteReadme();
	}
	
	private Element createSettingsElement(Document document, int mainx, int mainy, 
			boolean dieshow, boolean grayshow, boolean winshow) {
		Element element = document.createElement("settings");
		
		element.appendChild( createPositionElement(document, mainx, mainy) );
		element.appendChild( createDialogsElement(document, dieshow, grayshow, winshow) );
		
		return element;
	}
	
	private Element createDialogsElement(Document document, 
			boolean dieshow, boolean grayshow, boolean winshow) {
		Element element = document.createElement("dialogs");

		element.setAttribute("dieshow", String.valueOf(dieshow));
		element.setAttribute("grayshow", String.valueOf(grayshow));
		element.setAttribute("winshow", String.valueOf(winshow));
		
		return element;
	}
	
	private Element createPositionElement(Document document, int mainx, int mainy) {
		Element element = document.createElement("position");

		element.setAttribute("x", String.valueOf(mainx));
		element.setAttribute("y", String.valueOf(mainy));
		
		return element;
	}
	
}
