package org.amse.vbut.vzab.model;

import org.amse.vbut.vzab.*;

/**
 * Interface of the game
 * 
 * @author But Vasiliy
 *
 */
public interface IGame {

	static final int MIN_WIDTH  = 8;
	static final int MIN_HEIGHT = 8;
	
	static final int DEFAULT_WIDTH  = 12;
	static final int DEFAULT_HEIGHT = 12;

	static final int MAX_PLAYERS = 4;
	static final int MIN_PLAYERS = 2;
	
	static final int GRAYED_AREA_SIZE = 3;
	static final int MAX_GRAYED_MOVES = 4;

	/**
	 * Returns list of players.
	 * @return list of players.
	 */
	java.util.List<IPlayer> getPlayers();
	
	/**
	 * Sets specified object as list of players.
	 * @param players - list of players.
	 */
	void setPlayers(java.util.List<IPlayer> players);
	
	/**
	 * Returns current player number.
	 * @return current player number.
	 */
	int getCurrentPlayer();

	/**
	 * Sets player <tt>player</tt> as the curent player.
	 * @param player - number of player
	 */
	void setCurrentPlayer(int player);
	
	/**
	 * Return index of the specified player.
	 * @param player - player object
	 * @return	index of the player if specified player is in the game;
	 * 			<code>-1</code> otherwise.
	 */
	int getPlayerIndex(IPlayer player);
	
	/**
	 * @return current move index.
	 */
	int getCurrentMove();
	
	/**
	 * Sets current move index to movenumber value.
	 * @param movenumber - new current move value 
	 * @return	<tt>true</tt> if specified value is valid;
	 * 			<tt>false</tt> otherwise.
	 */
	boolean setCurrentMove(int movenumber);
	
	/**
	 * @return unmodifiable list of moves.
	 */
	java.util.List<IMove> getMoves();
	
	/**
	 * Method puts piece of the specified piece's player at the 
	 * specified position, checks if any enemy pieces were killed 
	 * and changes current player value.
	 * @param x - x-coordinate of the cell
	 * @param y - y-coordinate of the cell
	 * @param piece - parent piece of the new one
	 */
	void doMove(int x, int y, IPiece piece);

	/**
	 * Returns <tt>true</tt> if player can set his piece on 
	 * the specified cell and if the specified piece is the 
	 * parent piece of specified cell.
	 * 
	 * @param checkGrayedAread - specifies if method must
	 * check if grayed area full or not.
	 * 
	 * @return	<code>true</code> if player can do this move;
	 * 			<code>false</code> otherwise.
	 */
	boolean canDoMove(int x, int y, IPiece piece, boolean checkGrayedAread);
	
	/**
	 * Returns <tt>true</tt> if there are possible moves from
	 * this cell.
	 * @return <code>true</code> if there is piece on the 
	 * specified cell and this piece can be begining of the 
	 * current move.
	 */
	boolean canStartCurrentMove(int x, int y);
	
	/**
	 * Return <tt>true</tt> if there are possible moves from 
	 * the cell where the specified piece is.
	 * @return <code>true</code> if the specified piece can be 
	 * begining of the current move.
	 */
	boolean canStartCurrentMove(IPiece piece);
	
	/**
	 * Undoes move
	 */
	void undoMove();
	
	/**
	 * Returns <tt>true</tt> if it is possible to undo move.
	 * @return  <tt>true</tt> if it is possible to undo move
	 */
	boolean canUndoMove();
	
	/**
	 * Redoes move
	 */
	void redoMove();
	
	/**
	 * Returns <tt>true</tt> if it is possible to redo move.
	 * @return  <tt>true</tt> if it is possible to redo move;
	 * 			<tt>false</tt> otherwise.
	 */
	boolean canRedoMove();
	
	/**
	 * Returns board;
	 * @return board
	 */
	IBoard getBoard();
	
	/**
	 * Checks if the game is over.
	 * @return	<tt>true</tt> if the game is over; 
	 * 			<tt>false</tt> otherwise.
	 */
	boolean isGameOver();
	
	/**
	 * Returns base point for the specified player.
	 * @param playernum - index of player
	 * @return base point for the specified player.
	 */
	BoardPoint getPlayerBasePoint(int playernum);

	/**
	 * Returns parent point for the specified point 
	 * and the specified player.
	 * @param x x coordinate of the specified point
	 * @param y y coordinate of the specified point
	 * @param player specified player
	 * @return parent point for the specified point 
	 * if it exists, <tt>null</tt> otherwise.
	 */
	BoardPoint getParentPoint(int x, int y, IPlayer player);
	
	/**
	 * Returns modified-state of the game.
	 * @return if the game is modified.
	 */
	boolean isModified();
	
	/**
	 * Sets modified-state to the <i>modified</i> value.
	 * @param modified - new modified-state value
	 */
	void setModified(boolean modified);
	
	/**
	 * Returns <tt>true</tt> if specified player has been killed 
	 *  by last move.
	 * @param player - index of player
	 * @return <tt>true</tt> if specified player has been killed 
	 *  by last move, <tt>false</tt> otherwise.
	 */
	boolean isPlayerKilled(int player);

	/**
	 * Returns <tt>true</tt> if specified player is alive 
	 * @param player - index of player
	 * @return <tt>true</tt> if specified player is alive;
	 * <tt>false</tt> otherwise.
	 */
	boolean isPlayerAlive(int player);
	
	/**
	 * Returns rectangle that specifies grayed area around base point.
	 * @param point - base point
	 * @return rectangle that specifies grayed area around base point.
	 */
	BoardRectangle getGrayedRectangle(BoardPoint point);
	
	/**
	 * Returns <tt>true</tt> if grayed area of the specified player is full.
	 * @param player - index of player
	 * @return <tt>true</tt> if grayed area of the specified player is full
	 * <tt>false</tt> otherwise.
	 */
	boolean isGrayedAreaFull(int player);
	
}
