package org.amse.vbut.vzab.model.impl;

import org.amse.vbut.vzab.BoardPoint;
import org.amse.vbut.vzab.BoardRectangle;
import org.amse.vbut.vzab.VZabException;
import org.amse.vbut.vzab.model.*;

/**
 * Game class
 * 
 * @author But Vasiliy
 *
 */
class Game implements IGame {
	
	private final IBoard myBoard;
	
	private final int[] myGrayedMoves;
	private final int[] myPreviousGrayedMoves;
	
	private java.util.List<IPlayer> myPlayers;
	private boolean[] myPlayersAlive;
	private boolean[] myPlayersKilled;
	
	private int myCurrentPlayer;
	
	private java.util.Stack<IMove> myMoves;
	private int myCurrentMove;
	
	boolean myModified;
	
	/**
	 * Constructor of the Game
	 * @param w - width of the board
	 * @param h - height of the board
	 * @param players - array of players in the game
	 */
	public Game(int w, int h, java.util.List<IPlayer> players) {
		if (w < MIN_WIDTH || h < MIN_HEIGHT
				|| players.size() > MAX_PLAYERS
				|| players.size() < MIN_PLAYERS) {
			throw new VZabException("Wrong parameters in Game's constructor\n" 
					+ "w = " + w + "; h = " + h 
					+ "; players.length = " + players.size());
		}
		
		myPlayers = players;
		myPlayersAlive = new boolean[players.size()];
		myPlayersKilled = new boolean[players.size()];
		myGrayedMoves = new int[players.size()];
		myPreviousGrayedMoves = new int[players.size()];
		
		for (int i = 0; i < myPlayersAlive.length; ++i) {
			myPlayersAlive[i] = true;
			myPlayersKilled[i] = false;
			myGrayedMoves[i] = 0;
			myPreviousGrayedMoves[i] = 0;
		}
		
		myBoard = new Board(w, h);
		myCurrentPlayer = 0;
		
		myMoves = new java.util.Stack<IMove>();
		myCurrentMove = 0;
		
		myModified = false;
		
		setStartPosition();
	}
	
	public int getCurrentPlayer() {
		return myCurrentPlayer;
	}

	public void setCurrentPlayer(int player) {
		if (player < 0 || player >= getPlayers().size() || !myPlayersAlive[player]) {
			throw new VZabException("Bad player number in setCurrentPlayer method");
		}
		myCurrentPlayer = player;
		myModified = true;
	}

	public int getPlayerIndex(IPlayer player) {
		for (int i = 0; i < getPlayers().size(); ++i) {
			if (getPlayers().get(i) == player) {
				return i;
			}
		}
		return -1;
	}
	
	public BoardPoint getPlayerBasePoint(int playernum) {
		switch (playernum) {
		case 0: return new BoardPoint(0, 0);
		case 1: return new BoardPoint(getBoard().getWidth() - 1, getBoard().getHeight() - 1);
		case 2: return new BoardPoint(0, getBoard().getHeight() - 1);
		case 3: return new BoardPoint(getBoard().getWidth() - 1, 0);
		}
		return null;
	}
	
	private void setStartPosition()
	{
		final java.util.List<IPlayer> players = getPlayers();
		final int size = players.size();
		
		getBoard().clear();
		for (int i = 0; i < size; ++i) {
			final BoardPoint bp = getPlayerBasePoint(i);
			getBoard().setPiece(bp.getX(), bp.getY(), players.get(i));
		}
	}
	
	private void increaseCurrentPlayer() {
		final int playersnumber = getPlayers().size();
		final int oldplayer = getCurrentPlayer();
		int newplayer = (oldplayer + 1) % playersnumber;
		
		while (newplayer != oldplayer) {
			if (myPlayersAlive[newplayer]) {
				break;
			}
			newplayer = (newplayer + 1) % playersnumber;
		}
		
		setCurrentPlayer(newplayer);
	}
	
	private void decreaseCurrentPlayer() {
		final int playersnumber = getPlayers().size();
		final int oldplayer = getCurrentPlayer();
		int newplayer = oldplayer - 1;
		if (newplayer < 0) {
			newplayer = playersnumber - 1;
		}
		
		while (newplayer != oldplayer) {
			if (myPlayersAlive[newplayer]) {
				break;
			}
			if (--newplayer < 0) {
				newplayer = playersnumber - 1;
			}
		}
		
		setCurrentPlayer(newplayer);
	}
	
	private void updatePlayersAlive() {
		final int playersnumber = getPlayers().size();
		
		final IBoard board = getBoard();
		
		final int w = board.getWidth();
		final int h = board.getHeight();
		
		for (int i = 0; i < playersnumber; ++i) {
			myPlayersKilled[i] = myPlayersAlive[i];
			myPlayersAlive[i] = false;
		}
		
		for (int i = 0; i < w; ++i) {
			for (int j = 0; j < h; ++j) {
				final IPiece piece = board.getPiece(i, j);
				if (piece != null) {
					final IPlayer player = piece.getPlayer();
					final int index = getPlayerIndex(player);
					myPlayersAlive[index] |= canStartMove(i, j, piece); 
				}
			}
		}
		
		for (int i = 0; i < playersnumber; ++i) {
			BoardPoint bp = getPlayerBasePoint(i);
			if (bp == null) {
				myPlayersAlive[i] = false;
			} else {
				final int x = bp.getX();
				final int y = bp.getY();
				IPiece piece = getBoard().getPiece(x, y);
				
				if (piece == null) {
					myPlayersAlive[i] = false;
				} else {
					
					if (myPlayersAlive[i]) {
						myPlayersAlive[i] = piece.isAlive();
					} else {
						myPlayersAlive[i] = canStartMove(x, y, piece);
					}
					
				}
			}
		}
		
		for (int i = 0; i < playersnumber; ++i) {
			myPlayersKilled[i] ^= myPlayersAlive[i];
		}
		
	}

	private void updateKilled(java.util.List<IPiece> killed) {
		
		for (int i = 0; i < myPlayersKilled.length; ++i) {
			if (myPlayersKilled[i]) {
				
				final IBoard board = getBoard();
				final int width = board.getWidth();
				final int height = board.getHeight();
				final IPlayer player = getPlayers().get(i);
				
				for (int x = 0; x < width; ++x) {
					for (int y = 0; y < height; ++y) {
						IPiece piece = board.getPiece(x, y);
						
						if (piece != null 
								&& piece.getPlayer().equals(player) 
								&& piece.isAlive()) {
							piece.kill();
							killed.add(piece);
						}
					}
				}
				
			}
		}
		
	}
	
	private void reviveKilled(java.util.List<IPiece> killed) {
		for (IPiece p: killed) {
			p.revive();
		}
	}
	
	private boolean checkGrayedMoves(int x, int y, IPlayer player) {
		final int index = getPlayerIndex( player );
		final BoardPoint bp = getPlayerBasePoint(index);
		final BoardRectangle br = getGrayedRectangle(bp);
		
		if (br.inRect(x, y)) {
			if (myGrayedMoves[index] >= MAX_GRAYED_MOVES) {
				return false;
			}
		}
		
		return true;
	}

	private void increaseGrayedMoves(int x, int y, IPlayer player) {
		final int index = getPlayerIndex( player );
		final BoardPoint bp = getPlayerBasePoint(index);
		final BoardRectangle br = getGrayedRectangle(bp);
		
		if (br.inRect(x, y)) {
			if (myGrayedMoves[index] < MAX_GRAYED_MOVES) {
				++myGrayedMoves[index];				
				
				/*System.out.println("increaseGrayedMoves(): myPreviousGrayedMoves[" 
						+ index + "] = " + myPreviousGrayedMoves[index]);
				System.out.println("increaseGrayedMoves(): myGrayedMoves[" 
						+ index + "] = " + myGrayedMoves[index]);*/
				
			} else {
				throw new VZabException("Too many moves in grayed area of the player \"" + player.getName() + "\"");
			}
		}
	}
	
	private void decreaseGrayedMoves(int x, int y, IPlayer player) {
		final int index = getPlayerIndex( player );
		final BoardPoint bp = getPlayerBasePoint(index);
		final BoardRectangle br = getGrayedRectangle(bp);
		
		if (br.inRect(x, y)) {
			if (myGrayedMoves[index] > 0) {
				--myGrayedMoves[index];
				
				/*System.out.println("decreaseGrayedMoves(): myPreviousGrayedMoves[" 
						+ index + "] = " + myPreviousGrayedMoves[index]);
				System.out.println("decreaseGrayedMoves(): myGrayedMoves[" 
						+ index + "] = " + myGrayedMoves[index]);*/
				
			} else {
				throw new VZabException("Unable to decrease moves in grayed area of the player \"" + player.getName() + "\"");
			}
		}
	}
	
	public boolean canDoMove(int x, int y, IPiece piece, boolean checkGrayedAread) {
		if (!canStartCurrentMove(piece)) {
			return false;
		}
		
		if (x < 0 || x >= getBoard().getWidth() 
				|| y < 0 || y >= getBoard().getHeight()) {
			return false;
		}
		
		if (getBoard().getPiece(x, y) != null) {
			return false;
		}
		
		if (checkGrayedAread && !checkGrayedMoves(x, y, piece.getPlayer())) {
			return false;
		}
		
		return getBoard().getPiece(x - 1, y - 2) == piece
				|| getBoard().getPiece(x + 1, y - 2) == piece
				|| getBoard().getPiece(x - 1, y + 2) == piece
				|| getBoard().getPiece(x + 1, y + 2) == piece
				|| getBoard().getPiece(x - 2, y - 1) == piece
				|| getBoard().getPiece(x - 2, y + 1) == piece
				|| getBoard().getPiece(x + 2, y - 1) == piece
				|| getBoard().getPiece(x + 2, y + 1) == piece;
	}
	
	public void doMove(int x, int y, IPiece piece) {
		if (!canDoMove(x, y, piece, true)) {
			throw new VZabException("Can't do move");
		}
		
		//System.out.println("doMove(): currentplayer = " + getCurrentPlayer());

		getBoard().setPiece(x, y, piece.getPlayer());
		
		final int size = getPlayers().size();
		for (int i = 0; i < size; ++i) {
			myPreviousGrayedMoves[i] = myGrayedMoves[i];
		}
		
		increaseGrayedMoves(x, y, piece.getPlayer());
		
		java.util.List<IPiece> killed = killEnemies(x, y, piece.getPlayer());
		
		while (myCurrentMove < myMoves.size()) {
			myMoves.pop();
		}
		
		IMove move = new Move(x, y, piece, killed);
		
		myMoves.push(move);
		++myCurrentMove;
		
		updatePlayersAlive();
		updateKilled(move.getUpdatedKiled());
		increaseCurrentPlayer();
		
		myModified = true;
		
		//System.out.println("doMove(): currentplayer = " + getCurrentPlayer());
	}
	
	public boolean canUndoMove() {
		if (myMoves.empty() || myCurrentMove <= 0) {
			return false;
		}
		
		IMove move = myMoves.get(myCurrentMove - 1);
		
		if (getBoard().getPiece(move.getX(), move.getY()) == null) {
			return false;
		}
	
		return true;
	}

	public void undoMove() {
		if (myMoves.empty() || myCurrentMove <= 0) {
			throw new VZabException("Can't undo move");
		}
		
		//System.out.println("undoMove(): currentplayer = " + getCurrentPlayer());
		
		IMove move = myMoves.get(--myCurrentMove);
		
		if (getBoard().getPiece(move.getX(), move.getY()) == null) {
			
			++myCurrentMove;
			
			throw new VZabException("Can't undo move: there is no piece at the position:\n"
					+ "x = " + move.getX() + "; y = " + move.getY());
		}

		final int x = move.getX();
		final int y = move.getY();
		
		decreaseGrayedMoves(x, y, move.getPiece().getPlayer());
		getBoard().removePiece(x, y);
		
		reviveKilled(move.getUpdatedKiled());
		
		updatePlayersAlive();
		
		decreaseCurrentPlayer();
		
		reviveKilled(move.getKilled());

		updatePlayersAlive();
		
		myModified = true;

		//System.out.println("undoMove(): currentplayer = " + getCurrentPlayer());
	}
	
	public boolean canRedoMove() {
		if (myMoves.empty() || myCurrentMove >= myMoves.size()) {
			return false;
		}
		return true;
	}

	public void redoMove() {
		if (!canRedoMove()) {
			throw new VZabException("Can't redo move");
		}

		IMove move = myMoves.get(myCurrentMove);
		
		//System.out.println("redoMove(): currentplayer = " + getCurrentPlayer());
		
		final int x = move.getX();
		final int y = move.getY();
		final IPiece piece = move.getPiece();
		final IPlayer player = piece.getPlayer();

		if (!player.equals(getPlayers().get(getCurrentPlayer()))) {
			throw new VZabException("Error occured while undoing the move:\nplayer in move doesn't equals to the current player.");
		}

		getBoard().setPiece(x, y, player);
		increaseGrayedMoves(x, y, player);
		
		java.util.List<IPiece> killed = killEnemies(x, y, player);
		
		IMove newmove = new Move(x, y, move.getPiece(), killed);
		
		myMoves.set(myCurrentMove, newmove);
		
		++myCurrentMove;
		
		updatePlayersAlive();
		updateKilled(newmove.getUpdatedKiled());
		increaseCurrentPlayer();
		
		myModified = true;
		
		//System.out.println("redoMove(): currentplayer = " + getCurrentPlayer());
	}
	
	public IBoard getBoard() {
		return myBoard;
	}

	private boolean killEnemyPiece(int x, int y, IPlayer player,
			java.util.List<IPiece> killed) {
		IPiece p = getBoard().getPiece(x, y);
		if (p != null && !p.getPlayer().equals(player) && p.isAlive()) {
			p.kill();
			killed.add(p);
			return true;
		}
		return false;
	}

	private java.util.List<IPiece> killEnemies(int x, int y, IPlayer player) {
		java.util.List<IPiece> killed = new java.util.LinkedList<IPiece>();
		
		killEnemyPiece(x - 1, y - 1, player, killed);
		killEnemyPiece(x    , y - 1, player, killed);
		killEnemyPiece(x + 1, y - 1, player, killed);
		killEnemyPiece(x - 1, y    , player, killed);
		killEnemyPiece(x + 1, y    , player, killed);
		killEnemyPiece(x - 1, y + 1, player, killed);
		killEnemyPiece(x    , y + 1, player, killed);
		killEnemyPiece(x + 1, y + 1, player, killed);
		
		return killed;
	}
	
	private boolean isParentPoint(int x, int y, IPlayer player) {
		final IBoard board = getBoard();
		final IPiece piece = board.getPiece(x, y);
		if (piece == null || !piece.isAlive()) {
			return false;
		}
		if (piece.getPlayer().equals(player)) {
			return true;
		}
		return false;
	}
	
	public BoardPoint getParentPoint(int x, int y, IPlayer player) {
		if (isParentPoint(x - 2, y - 1, player)) {
			return new BoardPoint(x - 2, y - 1);
		}
		if (isParentPoint(x - 2, y + 1, player)) {
			return new BoardPoint(x - 2, y + 1);
		}
		if (isParentPoint(x + 2, y - 1, player)) {
			return new BoardPoint(x + 2, y - 1);
		}
		if (isParentPoint(x + 2, y + 1, player)) {
			return new BoardPoint(x + 2, y + 1);
		}
		if (isParentPoint(x - 1, y - 2, player)) {
			return new BoardPoint(x - 1, y - 2);
		}
		if (isParentPoint(x + 1, y - 2, player)) {
			return new BoardPoint(x + 1, y - 2);
		}
		if (isParentPoint(x - 1, y + 2, player)) {
			return new BoardPoint(x - 1, y + 2);
		}
		if (isParentPoint(x + 1, y + 2, player)) {
			return new BoardPoint(x + 1, y + 2);
		}
		return null;
	}

	/*
	 * Class Move defines one move in the game.
	 * 
	 * @author But Vasiliy
	 *
	 */
	private static class Move implements IMove{
		private final int myX, myY;
		private final IPiece myPiece;
		private final java.util.List<IPiece> myKilled;
		private final java.util.List<IPiece> myUpdatedKilled;

		public Move(int x, int y, IPiece piece, java.util.List<IPiece> killed) {
			myX = x;
			myY = y;
			myKilled = killed;
			myPiece = piece;
			myUpdatedKilled = new java.util.LinkedList<IPiece>();
		}

		public final int getX() {
			return myX;
		}

		public final int getY() {
			return myY;
		}

		public final java.util.List<IPiece> getKilled() {
			return myKilled;
		}
		
		public IPiece getPiece() {
			return myPiece;
		}
		
		public java.util.List<IPiece> getUpdatedKiled() {
			return myUpdatedKilled;
		}
	}
	
	private boolean isThereNoPiece(int x, int y) {
		if (x < 0 || x >= getBoard().getWidth() 
				|| y < 0 || y >= getBoard().getHeight()) {
			return false;
		}
		return getBoard().getPiece(x, y) == null;
	}
	
	private boolean canStartMove(int x, int y, IPiece piece) { 
		if (piece == null || !piece.isAlive()) {
			return false;
		}

		return isThereNoPiece(x - 1, y - 2)
				|| isThereNoPiece(x + 1, y - 2)
				|| isThereNoPiece(x - 1, y + 2)
				|| isThereNoPiece(x + 1, y + 2)
				|| isThereNoPiece(x - 2, y - 1)
				|| isThereNoPiece(x - 2, y + 1)
				|| isThereNoPiece(x + 2, y - 1)
				|| isThereNoPiece(x + 2, y + 1);
	}
	
	public boolean canStartCurrentMove(int x, int y) {
		IPiece piece = getBoard().getPiece(x, y);
		
		if (!canStartMove(x, y, piece)) {
			return false;
		}
		
		return piece.getPlayer() == getPlayers().get(getCurrentPlayer());
	}

	public boolean canStartCurrentMove(IPiece piece) {
		if (piece == null 
				|| !piece.isAlive() 
				|| piece.getPlayer() != getPlayers().get(getCurrentPlayer())) {
			return false;
		}
		
		final int width = getBoard().getWidth();
		final int height = getBoard().getHeight();
		
		for (int i = 0; i < width; ++i) {
			for (int j = 0; j < height; ++j) {
				final IPiece p = getBoard().getPiece(i, j);
				if (p != null && p.equals(piece)) {
					return canStartCurrentMove(i, j);
				}
			}
		}
		
		return false;
	}
	
	public boolean isGameOver() {
		
		final int playersnumber = getPlayers().size();
		int alivenum = 0;
		
		for (int i = 0; i < playersnumber; ++i) {
			if (myPlayersAlive[i]) {
				++alivenum;
			}
		}
		
		if (alivenum < IGame.MIN_PLAYERS) {
			return true;
		}
		
		for (int i = 0; i < getBoard().getWidth(); ++i) {
			for (int j = 0; j < getBoard().getHeight(); ++j) {
				if (canStartCurrentMove(i, j)) {
					return false;
				}
			}
		}
		
		return true;
	}

	public java.util.List<IPlayer> getPlayers() {
		return myPlayers;
	}

	public void setPlayers(java.util.List<IPlayer> players) {
		myPlayers = players;
		myModified = true;
	}

	public boolean setCurrentMove(int movenumber) {
		if (movenumber <= myMoves.size()) {
			myCurrentMove = movenumber;
			myModified = true;
			return true;
		}
		return false;
	}
	
	public int getCurrentMove() {
		return myCurrentMove;
	}

	public java.util.List<IMove> getMoves() {
		return java.util.Collections.unmodifiableList(myMoves);
	}

	public boolean isModified() {
		return myModified;
	}

	public void setModified(boolean modified) {
		if (modified == true) {
			throw new VZabException("Only Game can set its modified-state to <tt>true</tt>.");
		}
		myModified = modified;
	}

	public boolean isPlayerKilled(int player) {
		return myPlayersKilled[player];
	}

	public boolean isPlayerAlive(int player) {
		return myPlayersAlive[player];
	}

	public BoardRectangle getGrayedRectangle(BoardPoint point) {
		
		int x0 = point.getX() - GRAYED_AREA_SIZE;
		int y0 = point.getY() - GRAYED_AREA_SIZE;
		 
		int x1 = point.getX() + GRAYED_AREA_SIZE;
		int y1 = point.getY() + GRAYED_AREA_SIZE;
		
		final IBoard board = getBoard();
		final int width = board.getWidth();
		final int height = board.getHeight();
		
		if (x0 < 0) {
			x0 = 0;
		}
		if (y0 < 0) {
			y0 = 0;
		}
		if (x1 >= width) {
			x1 = width - 1;
		}
		if (y1 >= height) {
			y1 = height - 1;
		}

		return new BoardRectangle(x0, y0, x1, y1);
	}

	public boolean isGrayedAreaFull(int player) {
		return myGrayedMoves[player] == IGame.MAX_GRAYED_MOVES; 
			//&& myPreviousGrayedMoves[player] == (IGame.MAX_GRAYED_MOVES - 1);
	}

}
